package l12_generics;

/*
	Wildcards are represented by the question mark in Java "?" and they are used
	to referring to an unknown type. Wildcards are particularly useful when using
	generics and can be used as a parameter type but first, there is an
	important note to consider.

	It is known that Object is the supertype of all Java classes, however, a
	collection of Objects is not the supertype of any collection.
	For example, a List<Object> is not the supertype of List<String> and
	assigning a variable of type List<Object> to a variable of type List<String>
	will cause a compiler error. This is to prevent possible conflicts that can
	happen if we add heterogeneous types to the same collection.
*/

public class WildcardDemo {

	record Statistic<T extends Number>(T[] numbers) {

		//	Return type is double in all cases.
		public double average() {
			double sum = 0.0;
			for (T number : numbers) sum += number.doubleValue();
			return sum / numbers.length;
		}

		//	Determine if two averages are the same.
		//	Notice the use of the wildcard (Statistic with any type for T).
//		public boolean sameAvg(Statistic<?> statistic) {
		public boolean sameAvg(Statistic<Number> statistic) {
			return Math.abs(average() - statistic.average()) < 1e-5;
		}
	}

	public static void main(String[] args) {

//		Statistic<Integer> integerStatistic =
		Statistic<Number> integerStatistic =
			new Statistic<>(new Integer[] {1, 2, 3, 4, 5});
//		Statistic<Double> doubleStatistic =
		Statistic<Number> doubleStatistic =
			new Statistic<>(new Double[]{ 1.1, 2.2, 3.3, 4.4, 5.5});
//		Statistic<Float> floatStatistic =
		Statistic<Number> floatStatistic =
			new Statistic<>(new Float[] {1.0F, 2.0F, 3.0F, 4.0F, 5.0F});

		System.out.println("integerStatistic average is " + integerStatistic.average());
		System.out.println("doubleStatistic average is " + doubleStatistic.average());
		System.out.println("floatStatistic average is " + floatStatistic.average());
		System.out.println();

//	See which arrays have same average.
		System.out.print("Averages of integerStatistic and doubleStatistic ");
		System.out.println(integerStatistic.sameAvg(doubleStatistic) ?
				"are the same" : "differ");

		System.out.print("Averages of integerStatistic and floatStatistic ");
		System.out.println(integerStatistic.sameAvg(floatStatistic) ?
				"are the same" : "differ");
	}
}

